#include <linuxmt/types.h>
#include <linuxmt/config.h>
#include <linuxmt/sched.h>
#include <linuxmt/errno.h>
#include <linuxmt/serial_reg.h>

struct serial_info
{
	unsigned short io;
	unsigned char irq;
	unsigned char flags;
#define SERF_TYPE	15
#define SERF_EXIST	16
#define ST_8250		0
#define ST_16450	1
#define ST_16550	2
#define ST_UNKNOWN	15
};

static struct serial_info ports[4]=
{
	0x3f8,4,0,
	0x2f8,3,0,
	0x3e8,5,0,
	0x2e8,2,0
};

int rs_probe(sp)
struct serial_info *sp;
{
	int scratch,scratch2;
	int status1,status2;
	
	scratch=inb(sp->io+UART_IER);
	outb_p(0,sp->io+UART_IER);
	scratch=inb_p(sp->io+UART_IER);
	outb_p(scratch,sp->io+UART_IER);
	if(scratch2)
		return -1;
	
	scratch2=inb_p(sp->io+UART_LCR);
	outb(scratch2 | UART_LCR_DLAB, sp->io+UART_LCR);
	outb(0,sp->io+UART_EFR);
	outb(scratch2,sp->io+UART_LCR);
	outb(UART_FCR_ENABLE_FIFO, sp->io+UART_FCR);
	scratch=inb(sp->io+UART_IIR)>>6;
	switch(scratch)
	{
		case 0:
			sp->flags=SERF_EXIST|ST_16450;
			break;
		case 1:
			sp->flags=ST_UNKNOWN;
			break;
		case 2:
		case 3:	/* Could be a 16650.. we dont care */
			sp->flags=SERF_EXIST|ST_16550;
			break;
	}
	
	if(!scratch)
	{
		scratch=inb(sp->io+UART_SCR);
		outb(0xA5,sp->io+UART_SCR);
		status1=inb(sp->io+UART_SCR);
		outb(0x5A,sp->io+UART_SCR);
		status2=inb(sp->io+UART_SCR);
		outb(scratch,sp->io+UART_SCR);
		
		if((status1 != 0xA5) || (status2 != 0x5A))
			sp->io=ST_8250|SERF_EXIST;
	}
	
	/*
	 *	Reset the chip
	 */
	 
	outb(0x00,sp->io+UART_MCR);
	outb(UART_FCR_CLEAR_RCVR|UART_FCR_CLEAR_XMIT,sp->io+UART_FCR);
	inb(sp->io+UART_RX);
	return 0;
}

int rs_init()
{
	struct serial_info *sp=&ports[0];
	int i;
	
	printk("Serial driver version 0.01 with no serial options enabled\n");
	for(i=0;i<4;i++)
	{
		if(rs_probe(sp)==0)
		{
			printk("tty0%d at 0x%x (irq = %d)",
				i,sp->io,sp->irq);
			switch(sp->flags&SERF_TYPE)
			{
				case ST_8250:
					printk(" is a 8250\n");
					break;
				case ST_16450:
					printk(" is a 16450\n");
					break;
				case ST_16550:
					printk(" is a 16550\n");
					break;
				default:
					printk("\n");
					break;
			}
		}	
		sp++;
	}
	return 0;
}
