#include <linuxmt/config.h>
#include <linuxmt/errno.h>
#include <linuxmt/sched.h>

/*
 *	Find a free task slot.
 */

static int find_empty_process()
{
	int i;
	int unused=0;
	int n;
	
	for(i=0;i<MAX_TASKS;i++)
	{
		if(task[i].state==TASK_UNUSED)
		{
			unused++;
			n=i;
		}
	}
	
	if(unused==0 || (unused==1 && current->uid))
		return -EAGAIN;
	
	return n;
}


static int get_pid()
{
	static unsigned int last_pid = 0;
	struct task_struct *p;
	int i;
	
repeat:
	if(++last_pid==32768)
		last_pid=1;
			
	for(i=0;i<MAX_TASKS;i++)
	{
		p=&task[i];
		if(p->pid==last_pid || p->pgrp==last_pid || 
			p->session == last_pid)
			goto repeat;
	}
	return last_pid;
}

/*
 *	Clone a process.
 */
 
int do_fork()
{
	struct task_struct *t;
	int i=find_empty_process();
	if(i<0)
		return i;
	t=&task[i];
	
	/*
	 *	Copy registers 
	 */
	 
	t->t_regs=current->t_regs;
	t->t_regs.ksp=&t->t_kstack+126;
	
	/*
	 *	Copy CPU state.
	 */
	 
	t->state = TASK_UNINTERRUPTIBLE;
	t->pid = get_pid();
	t->uid = current->uid;
	t->euid = current->euid;
	t->suid = current->suid;
	t->gid = current->gid;
	t->egid = current->egid;
	t->sgid = current->sgid;
	t->session = current->session;
	t->ppid = current->pid;
	
	/*
	 *	Build a return stack for t.
	 */
	 
	arch_build_stack(t);
	
	/*
	 *	Wake our new process
	 */
	 
	wake_up_process(t);
	
	/*
	 *	Return the created task.
	 */
	
	return t->pid;
}

